import vf;
import pytest;

class Subint (int): pass;
class Sublist (list): pass;
class Subdict (dict): pass;

eg_subint = Subint(100);
eg_sublist = Sublist([1, "a", True]);
eg_subdict = Subdict({"foo": "bar"});

def test_typeIs ():
    assert vf.typeIs(bool)(True) == True;
    assert vf.typeIs(int)(1) == True;
    assert vf.typeIs(int)("one") == False;
    assert vf.typeIs(float)(3.33) == True;
    assert vf.typeIs(str)("foo") == True;
    assert vf.typeIs(list)([]) == True;
    assert vf.typeIs(list)(eg_sublist) == False;
    assert vf.typeIs(dict)({"foo":" bar"}) == True;
    assert vf.typeIs(dict)(eg_subdict) == False;
    assert vf.typeIs(Subdict)(eg_subdict) == True;
    assert vf.typeIs(Subdict)({}) == False;

def test_instanceOf ():
    assert vf.instanceOf(int)(1) == True;
    assert vf.instanceOf(int)(eg_subint) == True;
    assert vf.instanceOf(dict)({}) == True;
    assert vf.instanceOf(dict)(eg_subdict) == True;
    assert vf.instanceOf(str)({}) == False;
    assert vf.instanceOf(str)(100) == False;
    
    assert vf.instanceOf(int, str)(1) == True;
    assert vf.instanceOf(int, str)("one") == True;
    assert vf.instanceOf(int, str)({}) == False;
    assert vf.instanceOf(int, dict)({}) == True;
    assert vf.instanceOf(int, dict)(eg_subdict) == True;

def test_typeIn ():
    assert vf.typeIn(int, dict)({}) == True;
    assert vf.typeIn(int, dict)(eg_subdict) == False;

def test_dictOf_basicPostSchema ():
    validator1 = vf.dictOf({
        "_id": vf.typeIs(str),
        "title": vf.typeIs(str),
        "body": vf.typeIs(str),
    });
    assert validator1({"_id": "i", "title": "t", "body": "b"});
    assert validator1({"_id": "", "title": "", "body": ""});
    
    with pytest.raises(vf.ValidationError) as e:
        validator1("Foo!");
    assert "Expected dict-like object" in e.value.args[0];
    
    with pytest.raises(vf.ValidationError) as e:
        validator1({"_id": "a", "title": "t"});
    assert "missing required keys: {'body'}" in e.value.args[0];
    
    with pytest.raises(vf.ValidationError) as e:
        validator1({"_id": "", "title": "", "body": "", "xtra": ""});
    assert "excess keys: {'xtra'}" in e.value.args[0];
    
    
    validator2 = vf.dictOf({
        "_id":   vf.allOf(vf.truthy, vf.typeIs(str)),
        "title": vf.allOf(vf.truthy, vf.typeIs(str)),
        "body":  vf.allOf(vf.truthy, vf.typeIs(str)),
    });
    assert validator1({"_id": "a", "title": "b", "body": "c"});
    with pytest.raises(vf.ValidationError) as e:
        validator2({"_id": "", "title": "b", "body": "c"});
    assert ("key: %r" % "_id") in e.value.args[0];

def test_wrt_readme ():
    personValidator = vf.dictOf({
        "_id": vf.typeIs(str),
        "name": vf.dictOf({
            "first": vf.typeIs(str),            # ---- Equivalent
            "last": lambda x: type(x) is str,   # -----^^^
        }),
        "birth_timestamp": vf.typeIs(int),
        "gender": lambda x: x in ["MALE", "FEMALE", "OTHER"],
        "parentIdList": vf.listOf(vf.typeIs(str)),
        "spouceId": lambda x: x is None or type(x) is str,
        "childrenIdList": vf.listOf(vf.typeIs(str)),
    });
    personValidator({
        "_id": "00a3",
        "name": {"first": "John", "last": "Doe"},
        "birth_timestamp": 318191400000,
        "gender": "MALE",
        "parentIdList": ["00a1", "00a2"],
        "spouceId": "87b1",
        "childrenIdList": ["00a6", "00a8"],
    });
    with pytest.raises(vf.ValidationError) as e:
        personValidator({
            "_id": "1c3f",
            "name": {"first": "Jane", "last": "Doe"},
            "birth_timestamp": "1990-01-01",
            "gender": "FEMALE",
            "parentIdList": ["1c3a", "1c3b"],
            "spouceId": None,
            "childrenIdList": [],
        });
    assert "key: 'birth_timestamp'" in e.value.args[0];

# End ######################################################



"""
VF: Validation Functions (for Python dicts.)

Copyright (c) 2020 Polydojo, Inc.

SOFTWARE LICENSING
------------------
The software is released "AS IS" under the MIT License,
WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED. Kindly
see LICENSE.txt for more details.

NO TRADEMARK RIGHTS
-------------------
The above software licensing terms DO NOT grant any right in the
trademarks, service marks, brand names or logos of Polydojo, Inc.
""";

import functools;
import re;

__version__ = "0.0.1-preview";  # Req'd by flit.

identity = lambda x: x;
truthy = lambda x: bool(x);
falsy = lambda x: not x;
nonely = lambda x: x is None;   # <-- TODO: Rename?

def typeIs (typ):
    "Makes `func (x)` for checking if `type(x) is typ`.";
    return lambda x: type(x) is typ;

def instanceOf (*typTup):
    "Makes `func (x)` for checking if `isinstance(x, typTup)`.";
    return lambda x: isinstance(x, typTup);

def typeIn (*typTup):
    "Makes `func (x)` for checking if `type(x) in typTup`.";
    return lambda x: type(x) in typTup;

def reMatch (pattern):
    "Makes `func (s)` for checking `s` against `pattern`.";
    if type(pattern) is str:
        return lambda s: re.match(pattern, s);
    if type(pattern) is re.Pattern:
        return lambda s: pattern.match(s);
    raise ValueError("Expected `pattern` to be of type "
        "`str` or `re.Pattern`, not: %r" % (pattern,)
    );

def allOf (*fns):
    "Makes `func (x)` for checking `all(fn(x) for fn in fns)`.";
    return lambda x: all(map(lambda fn: fn(x), fns));

def anyOf (*fns):
    "Makes `func (x)` for checking `any(fn(x) for fn in fns)`.";
    return lambda x: any(map(lambda fn: fn(x), fns));

def listOf (fn):
    "Makes `func (li)` for checking `all(fn(x) for x in li`).";
    return lambda li: isinstance(li, list) and all(map(fn, li));

class BadSchemaError (ValueError): pass;
class ValidationError (ValueError): pass;

def _validateSchemaItself (schema):
    "Ensures that `schema` itself is valid.";
    if not isinstance(schema, dict):
        raise BadSchemaError("Not an instance of `dict`.");
    for key, rhsFn in schema.items():
        if not callable(rhsFn):
            raise BadSchemaError(
                "Non-callable value against key: %r" % (key,),
            );
    return True;

def dictOf (schema, extraKeysOk=False):
    "Makes `func (d)` for _validating_ `d` agaisnt `schema`.";
    assert _validateSchemaItself(schema);
    def validateFn (d):
        if not isinstance(d, dict):
            raise ValidationError(
                "Expected dict-like object, not: %r" % (d,),
            );
        dKeySet = set(d.keys());
        sKeySet = set(schema.keys());
        if not dKeySet.issuperset(sKeySet):
            raise ValidationError("Dict-like object is missing " +
                "required keys: {}".format(sKeySet - dKeySet) #+
            );
        if (not extraKeysOk) and (dKeySet != sKeySet):
            raise ValidationError("Dict-like object has " +
                "excess keys: {}".format(dKeySet - sKeySet) #+
            );
        for key, rhsFn in schema.items():
            assert callable(rhsFn);
            if not rhsFn(d[key]):
                raise ValidationError(
                    ("Against key: %r\n" % (key,)) +
                    ("Unexpected value: %r" % (d[key],)) #+
                );
        return True;
    return validateFn;

# End ######################################################



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup


setup(name='vf',
      version='0.0.1rc0',
      description='VF: Validation Functions (for Python dicts.)',
      author='Polydojo, Inc.',
      author_email='8079593+sumukhbarve@users.noreply.github.com',
      url='https://github.com/polydojo/vf',
      py_modules=['vf'],
     )
